/*
  Name:        	Do178Builder-utility.h
  Author:      	Ron Burkey
  Purpose:     	Header file for the Do178Builder-utility.c file and
  		friends.
  Mod History:	03/13/2008 RSB	Began.
  		03/17/2008 RSB	Added some stuff related to --trace-verify.
		03/21/2008 RSB	Added section titles.
  
  Copyright 2008 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef DO178BUILDER_UTILITY_H
#define DO178BUILDER_UTILITY_H

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

//--------------------------------------------------------------------
// Constants.

#define COPYRIGHT "2008"
#define MAX_SECTION_DEPTH 30
// The following is based on the notion that the largest possible
// identifier would be a section number consisting of the 
// 999.999.999....999 (at the maximum section depth).  (The NUL terminator 
// for the string uses up one dot position.)  This seems like a bit of
// overkill, and should be fine for any purpose.
#define MAX_REQUIREMENT_SIZE (4 * MAX_SECTION_DEPTH)
#define SHORT_REQUIREMENT_SIZE 21
// Limits on the number of requirements in any given node.
#define MAX_MATCHES_PER_NODE 500
#define MAX_ONE_TO_ONE 5

// A regular expression for parsing stuff such as
//	<docbook>CreamyCenter</docbook>
// to find the creamy center.  Interestingly, even though the characters
// '<' and '>' are represented in the project database XML as "&lt;" and 
// "&gt;", libxml apparently converts them back.  The middle portion --- 
// any string not containing an ampersand --- is in parantheses to 
// allow it to allow it to be presented to us as a separate field in the 
// regex match (without the opening and closing tags).
#define SYSTEM_LIST_PATTERN \
  "<[tT][rR][aA][cC][eE]>" \
  "([^&]*)" \
  "</[tT][rR][aA][cC][eE]>"

//--------------------------------------------------------------------
// Datatypes.

enum { REQ_DOC, DES_DOC, VER_DOC };

// Structures for tracking section numbers of requirements, design, and
// verification documents.
typedef struct {
  int WalkLevel;
  int Present;
  int Depth;
  int Levels[MAX_SECTION_DEPTH];
  // The following is an ASCII-string version of the document number
  // given by the Levels array. 
  char Number[MAX_REQUIREMENT_SIZE];
  char *Title;
} SectionNumber_t;

typedef SectionNumber_t DocSections_t[3];

// Structures for creating traceability matrices.
typedef struct {
  SectionNumber_t HighLevelRequirementSection;
  char HighLevelRequirements[MAX_MATCHES_PER_NODE][MAX_REQUIREMENT_SIZE];
  int NumHighLevelRequirements;
  int LowLevelCount;
} TraceDesignState_t;

typedef struct {
  SectionNumber_t HighLevelRequirementSection;
  char HighLevelRequirements[MAX_MATCHES_PER_NODE][MAX_REQUIREMENT_SIZE];
  int NumHighLevelRequirements;
  int LowLevelCount;
  int LowSectionCount;
} TraceDesignFullState_t;

typedef struct {
  SectionNumber_t SystemRequirementSection;
  char SystemRequirements[MAX_MATCHES_PER_NODE][MAX_REQUIREMENT_SIZE];
  char OneToOne[MAX_MATCHES_PER_NODE][MAX_ONE_TO_ONE][SHORT_REQUIREMENT_SIZE];
  int NumOneToOne[MAX_MATCHES_PER_NODE];
  int NumSystemRequirements;
  int HighLevelCount;
  char HighLevelRequirements[MAX_MATCHES_PER_NODE][MAX_REQUIREMENT_SIZE];
  int NumHighLevelRequirements;
} TraceSystemState_t;

typedef struct {
  // By a "requirement", we mean either a high-level requirement from 
  // the SRD/HRD or a low-level requirement from the SDD/HDRD.
  SectionNumber_t RequirementSection;
  char Requirements[MAX_MATCHES_PER_NODE][MAX_REQUIREMENT_SIZE];
  int NumRequirements;
  char DocName[33];	// From which the requirement came.
  int TestCaseCount;
} TraceVerifyState_t;

// Holds a set of options for processing.
typedef struct {

  // From command line switches, general.
  char *DatabaseName;	// Name of the input file.
  int Display;		// Display a simple tag map of the input file.
  int NoSed;		// Don't translate entities in the input file.
  int CombineSrdSdd;
  int Sortable;		// Stringify in a way that's more easily sortable.
  // Command line switches to activate various trace data.
  int TraceDesign;
  int TraceDesignFull;
  int TraceSystem;
  int TraceImplement;
  int TraceVerify;
  // Command line switches for various regular-expression patterns.
  char *HighLabelPattern;
  char *HighSectionPattern;
  char *HighSectionRange;
  char *LowLabelPattern;
  char *LowSectionPattern;
  char *LowSectionRange;
  char *VerifySectionRange;
  char *SystemSectionRange;
  char *SystemListPattern;
  // Command-line switches for regexp properties.
  int RegexCaseInsensitive;
  int RegexBasic;	
  int RegexExtended;	// Now derived from RegexBasic.
  // Command-line switches for trace-field delimiters within 
  // <trace></trace> blocks.
  char DelimiterSystem;
  char DelimiterImplement;
  
  // Names of the various documents.  (Either SRD/SDD/SVCP, or else
  // HRD/HDRD/HTPR.)
  char *ReqDocName;	// SRD or HRD
  char *DesDocName;	// SDD or HDRD
  char *VerDocName;	// SVCP or HTPR
  char *ComDocName;	// SRD/SDD or HRD/HDRD
  
  // Compiled state machines for the regex patterns above.
  regex_t HighLabelPatternMachine;
  regex_t HighSectionPatternMachine;
  regex_t HighSectionRangeMachine;
  regex_t LowLabelPatternMachine;
  regex_t LowSectionPatternMachine;
  regex_t LowSectionRangeMachine;
  regex_t VerifySectionRangeMachine;
  regex_t SystemSectionRangeMachine;
  regex_t SystemListPatternMachine;
  
  // For processing.
  int WalkLevel;
  DocSections_t DocSections;
  TraceDesignState_t TraceDesignState;
  TraceDesignFullState_t TraceDesignFullState;
  TraceVerifyState_t TraceVerifyState, TraceVerifyStateParent;
  int VerifyChildWalkLevel, VerifyParentWalkLevel;
  TraceSystemState_t TraceSystemState;
  
  // Output files for trace data.
  FILE *FileTraceDesign;
  FILE *FileTraceDesignFull;
  FILE *FileTraceSystem;
  FILE *FileTraceImplement;
  FILE *FileTraceVerify;
  
} ProcessingOptions_t;

// Returns 0 on success, non-zero on error.
typedef int NodeFunction_t (xmlNode *a_node, ProcessingOptions_t *Options);

//--------------------------------------------------------------------
// Global variables.

#ifdef IS_DO178BUILDER_UTILITY_MAIN
regmatch_t PatternMatches[MAX_MATCHES_PER_NODE];
#else
extern regmatch_t PatternMatches[MAX_MATCHES_PER_NODE];
#endif

//--------------------------------------------------------------------
// Function prototypes.

void StringifySectionNumber (SectionNumber_t *Section, ProcessingOptions_t *Options);
char *Outputtable (SectionNumber_t *Section, ProcessingOptions_t *Options);
int CompareSectionNumbers (SectionNumber_t *Section1, SectionNumber_t *Section2);
int WalkNodes (xmlNode * a_node, NodeFunction_t *f, ProcessingOptions_t *Options);
NodeFunction_t PrintNodeName;
NodeFunction_t Trace;
NodeFunction_t TraceDesign;
NodeFunction_t TraceDesignFull;
NodeFunction_t TraceSystem;
NodeFunction_t TraceImplement;
NodeFunction_t TraceVerify;
void TraceInitialize (ProcessingOptions_t *Options);
void TraceCloseout (ProcessingOptions_t *Options);
int Pattern (regex_t *Machine, const char *PatternString, ProcessingOptions_t *Options);
void PrettifyTitle (const char *Raw, char *Pretty);


#endif // DO178BUILDER_UTILITY_H

