/*
  Name:        	TraceDesign.c
  Author:      	Ron Burkey
  Purpose:     	The processing for Do178Builder-utility 
  		that creates the traceability matrix for the
		--trace-design command-line switch.
  Mod History:	03/14/2008 RSB	Began.
  		03/16/2008 RSB	Began incorporating features
				for TraceDesignFull.c.
		03/17/2008 RSB	Fixed a bug related to design-full,
				in which the high-level requirements
				might not be listed if there were
				no low-level requirements.
  
  Copyright 2008 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
  
  The dataset output by this function is a CSV file with one record 
  for each pairing of a high-level requirement with a low-level 
  requirement.  A requirement which has no association is matched
  with "(none)".  For example, suppose that high-level requirement
  H1 is associated with low-level requirements L1 and L2, low-level 
  requirement L3 is matched with high-level requirements H2 and H3
  (this shouldn't be possible!), H4 is matched with L4, H5 is 
  matched with nothing, and L5 is matched with nothing.  (The actual
  names of the requirements are determined by user-supplied information
  and are unlike to be something as simple as H5 or L5.  For example,
  the high-level requirements might be things like EASDD.1234, and the
  low-level requirements might be things like [SL5432].  I merely use
  things like H5 and L5 because they're simple to write.)  Then the
  output file would look like the following, though not necessarily
  with the records in the same order (with the first line containing
  column headers):
  
  	"High-Level Requirement","Low-Level Requirement"
  	"SRD H1","SDD L1"
	"SRD H1","SDD L2"
	"SRD H2","SDD L3"
	"SRD H3","SDD L3"
	"SRD H4","SDD L4"
	"SRD H5","(none)"
	"(none)","SDD L5"

  You might think that the notations "SRD" and "SDD" above are pointless
  (e.g., why "SRD H1" rather than just "H1"?), but reason they are 
  there is to allow easy combination of CSV data created from 
  different databases into a single table.  For example, one could 
  combine data from a DO-178B database and a DO-254 database, and
  the face that some items read "SRD H1" while others read "HRD H1"
  would allow the requirements from the different databases to 
  remain distinct.

  The requirements are identified by pattern matching with user-
  supplied regular expressions.  The associations occur whenever
  the low-level requirement appears either within the same node
  (document section) as the high-level requirement, or in any 
  child node (document sub-section) of that node.
  
  By sucking the CSV file into a spreadsheet program, it can be sorted 
  in order of the high-level requirements or in order of the low-level
  requirements.  The actual output order is simply the order in which
  the requirements are found in the XML.
*/


///////////////////////////////////////////////////////////////////////
//                  *****  IMPORTANT NOTE!  *****

// This file was written for the --trace-design command-line switch,
// but with mimimal changes supports --trace-design-full as well.
// For that alternate support, TRACE_DESIGN_FULL must be #defined.
// In other words, this file is compiled twice, once with 
// TRACE_DESIGN_FULL, and once with it.

//               *****  END OF IMPORTANT NOTE!  *****
///////////////////////////////////////////////////////////////////////


#include "Do178Builder-utility.h"
#include <ctype.h>

// This must *follow* the #include for Do178Builder-utility.h, or else
// it will create duplicate names within that file.
#ifndef TRACE_DESIGN_FULL
static char *CsvFilename = "Do178Builder-trace-design.csv";
#else
static char *CsvFilename = "Do178Builder-trace-design-full.csv";
// I got the following list simply by searching for the string "Design".
#define TraceDesignState TraceDesignFullState
#define TraceDesignCloseout TraceDesignFullCloseout
#define FileTraceDesign FileTraceDesignFull
#define TraceDesign TraceDesignFull
#endif // TRACE_DESIGN_FULL


// Close out any high-level collection.
void
TraceDesignCloseout (ProcessingOptions_t *Options)
{
  int k;
  char *ReqDocName;
  
  if (Options->CombineSrdSdd)
    {
      ReqDocName = Options->ComDocName;
    }
  else
    {
      ReqDocName = Options->ReqDocName;
    }
    
#ifdef TRACE_DESIGN_FULL
  Options->TraceDesignState.LowLevelCount += Options->TraceDesignState.LowSectionCount;
  Options->TraceDesignState.LowSectionCount = 0;
#endif    
  if (!Options->TraceDesignState.LowLevelCount)
    {
      for (k = 0; k < Options->TraceDesignState.NumHighLevelRequirements; k++)
        {
#ifndef TRACE_DESIGN_FULL	
	  fprintf (Options->FileTraceDesign, "\"%s %s\",\"%s\"\n",
		   ReqDocName,
		   Options->TraceDesignState.HighLevelRequirements[k],
		   "(none)");
#else
	  fprintf (Options->FileTraceDesign, "\"%s %s\",\"%s %s\",\"%s\",\"%s\"\n",
		   ReqDocName,
		   Options->TraceDesignState.HighLevelRequirements[k],
		   ReqDocName,
		   Options->TraceDesignState.HighLevelRequirementSection.Number,
		   "(none)","(none)");
#endif		   
        }
    }
  Options->TraceDesignState.LowLevelCount = 0;
  Options->TraceDesignState.NumHighLevelRequirements = 0;
}

// This function does the processing by being called by WalkNodes
// once for each node in the XML tree.  It has to maintain its own
// state machine in order to tie all of these separate calls together,
// and the state machine comprises fields in Options.  Calling with
// a_node==NULL should initialize (reset) the state machine.
int 
TraceDesign (xmlNode *a_node, ProcessingOptions_t *Options)
{
  int i, j, k, n;
  char c;
  int CanContainHighLevel = 0, CanContainLowLevel = 0;
  int DoesContainHighLevel = 0, DoesContainLowLevel = 0;
  regmatch_t pmatch[1];
  char *Text, *ReqDocName, *DesDocName;
  //char *s;
      
  // Initialize?
  if (a_node == NULL)
    {
      Options->FileTraceDesign = fopen (CsvFilename, "w");
      if (Options->FileTraceDesign == NULL)
        {
	  fprintf (stderr, "Error: Cannot create %s\n", CsvFilename);
	  return (1);
	}
      Options->TraceDesignState.NumHighLevelRequirements = 0;
#ifndef TRACE_DESIGN_FULL
      fprintf (Options->FileTraceDesign, 
               "\"High-Level Requirement\",\"Low-Level Requirement\"\n");
#else
      fprintf (Options->FileTraceDesign, 
               "\"High-Level Requirement\","
	       "\"Requirements Document Section\","
	       "\"Low-Level Requirement\","
	       "\"Design Document Section\"\n");
#endif
      return (0);
    }
    
  if (Options->CombineSrdSdd)
    {
      ReqDocName = DesDocName = Options->ComDocName;
    }
  else
    {
      ReqDocName = Options->ReqDocName;
      if (!strcmp (Options->DesDocName, "DD"))
        DesDocName = "SDD";
      else
        DesDocName = Options->DesDocName;
    }
    
  // Process the node. ---------------------------------------------
  
  // Determine if it can contain a high-level requirement.  (Not 
  // necessarily that it DOES, just that it COULD.)  This will
  // be so if:
  //	1. It is part of the requirements document; and
  //	2. It matches the section-number constraint from the command
  //	   line switches.
  // By default, it is marked as NOT possible, so we only need to 
  // check the positive cases.
  if (Options->DocSections[REQ_DOC].Present)
    {
      if (Options->HighSectionRange == NULL)
        CanContainHighLevel = 1;	// No range constraint, so all okay!
      else
        {
	  // A section-number pattern has been defined by a command-line
	  // switch as a constraint on the sections that can contain
	  // requirements, so apply it and see if we're in the right
	  // range of section numbers.  Note that we make no effort
	  // to insure that the match encompasses the entire section
	  // number.  If the user desires that, then he should have
	  // prefixed and suffixed his pattern with '^' and '$' to 
	  // match the beginning of line and end of line.  If regexec
	  // returns 0, then it means a match was found.  
	  if (0 == regexec (&Options->HighSectionRangeMachine,
	  		    Options->DocSections[REQ_DOC].Number,
			    1, pmatch, 0))
	    CanContainHighLevel = 1;
	}
    }

  // If we've found a section that could contain high-level requirements,
  // then we need to close out any existing high-level requirements that 
  // we may have been accumulating low-level requirements for.
  if (CanContainHighLevel)
    TraceDesignCloseout (Options);

  // If could contain high-level requirements, let's check to see if it
  // DOES contain any.  If so, we have to buffer them.  We can't do anything
  // with them immediately, since we can only write them out when we have
  // some low-level requirements to go with them.
  if (CanContainHighLevel)
    {
      Options->TraceDesignState.HighLevelRequirementSection = Options->DocSections[REQ_DOC];
      // Get the text of the node.
      Text = (char *) xmlNodeListGetString (a_node->doc, a_node->xmlChildrenNode, 1);
      //for (s = Text; *s; s++)
      //  if (isspace (*s))
      //    *s = ' ';
      //if (xmlStrlen (Text) > 50)
      //  strcpy ((char *) &Text[47], "...");
      printf("HL candidate %s, section %s:", Options->ReqDocName, 
             Options->DocSections[REQ_DOC].Number);
      
      // Now determine if there are any matches for the requirements
      // language patterns.  We only check the so-called "section" language
      // matches if there are no "label" matches.
      for (i = j = 0; !regexec (&Options->HighLabelPatternMachine, &Text[j], 1, pmatch, 0); i++, j = pmatch[0].rm_eo)
        {
	  pmatch[0].rm_so += j;		// Turn relative offsets into absolute.
	  pmatch[0].rm_eo += j;
	  n = pmatch[0].rm_eo - pmatch[0].rm_so;
	  if (n >= MAX_REQUIREMENT_SIZE)
	    {
	      fprintf (stderr, "Error: Pattern produces string length >%d\n", MAX_REQUIREMENT_SIZE - 1);
	      xmlFree (Text);
	      return (1);
	    }
	  if (i >= MAX_MATCHES_PER_NODE)
	    {
	      fprintf (stderr, "Error: Too many requirements in a section.\n");
	      xmlFree (Text);
	      return (1);
	    }
	  c = Text[pmatch[0].rm_eo];
	  Text[pmatch[0].rm_eo] = 0;
	  strcpy (Options->TraceDesignState.HighLevelRequirements[i], &Text[pmatch[0].rm_so]);
	  Text[pmatch[0].rm_eo] = c;
	}
      for (; !i && !regexec (&Options->HighSectionPatternMachine, Text, 1, pmatch, 0); i++)
        {
	  if (i >= MAX_MATCHES_PER_NODE)
	    {
	      fprintf (stderr, "Error: Too many requirements in a section.\n");
	      xmlFree (Text);
	      return (1);
	    }
	  strcpy (Options->TraceDesignState.HighLevelRequirements[i], 
	          Options->DocSections[REQ_DOC].Number);
	}
      Options->TraceDesignState.NumHighLevelRequirements = i;
      if (i == 0)
        {
          printf (" (none)");
#ifdef TRACE_DESIGN_FULL
	  fprintf (Options->FileTraceDesign, "\"%s\",\"%s %s\",\"%s\",\"%s\"\n",
		   "(none)",
		   ReqDocName,
		   Options->TraceDesignState.HighLevelRequirementSection.Number,
		   "(none)", "(none)");
#endif	  
	}
      else
	for (i = 0; i < Options->TraceDesignState.NumHighLevelRequirements; i++)
	  printf (" %s", Options->TraceDesignState.HighLevelRequirements[i]);
      printf ("\n");
      xmlFree (Text);      
    }

  // Similarly, for the low-level requirements.  
  if (Options->DocSections[DES_DOC].Present)
    {
      if (Options->LowSectionRange == NULL)
        CanContainLowLevel = 1;	// No range constraint, so all okay!
      else
        {
	  if (0 == regexec (&Options->LowSectionRangeMachine,
	  		    Options->DocSections[DES_DOC].Number,
			    1, pmatch, 0))
	    CanContainLowLevel = 1;
	}
    }
    
  if (CanContainLowLevel)
    {
      char *s;
#ifdef TRACE_DESIGN_FULL      
      Options->TraceDesignState.LowSectionCount++;
#endif      
      // Get the text of the node.
      Text = (char *) xmlNodeListGetString (a_node->doc, a_node->xmlChildrenNode, 1);
      //for (s = Text; *s; s++)
      //  if (isspace (*s))
      //    *s = ' ';
      //if (xmlStrlen (Text) > 50)
      //  strcpy ((char *) &Text[47], "...");
      printf("LL candidate %s, section %s:", Options->DesDocName, 
             Options->DocSections[DES_DOC].Number, Text);
      
      // Now determine if there are any matches for the requirements
      // language patterns.  We only check the so-called "section" language
      // matches if there are no "label" matches.
      for (i = j = 0; !regexec (&Options->LowLabelPatternMachine, &Text[j], 1, pmatch, 0); i++, j = pmatch[0].rm_eo)
        {
	  pmatch[0].rm_so += j;		// Turn relative offsets into absolute.
	  pmatch[0].rm_eo += j;
	  n = pmatch[0].rm_eo - pmatch[0].rm_so;
	  c = Text[pmatch[0].rm_eo];
	  Text[pmatch[0].rm_eo] = 0;
	  printf (" %s", &Text[pmatch[0].rm_so]);
	  for (k = 0; k < Options->TraceDesignState.NumHighLevelRequirements; k++)
	    {
#ifndef TRACE_DESIGN_FULL
	      fprintf (Options->FileTraceDesign, "\"%s %s\",\"%s %s\"\n",
		       ReqDocName,
		       Options->TraceDesignState.HighLevelRequirements[k],
		       DesDocName,
		       &Text[pmatch[0].rm_so]);
#else
	      fprintf (Options->FileTraceDesign, "\"%s %s\",\"%s %s\",\"%s %s\",\"%s %s\"\n",
		       ReqDocName,
		       Options->TraceDesignState.HighLevelRequirements[k],
		       ReqDocName,
		       Options->TraceDesignState.HighLevelRequirementSection.Number,
		       DesDocName,
		       &Text[pmatch[0].rm_so],
		       DesDocName,
		       Options->DocSections[DES_DOC].Number);
#endif		       
            }		     
	  if (k == 0)
	    {
#ifndef TRACE_DESIGN_FULL
	      fprintf (Options->FileTraceDesign, "\"%s\",\"%s %s\"\n",
		       "(none)",
		       DesDocName,
		       &Text[pmatch[0].rm_so]);
#else		       
	      fprintf (Options->FileTraceDesign, "\"%s\",\"%s %s\",\"%s %s\",\"%s %s\"\n",
		       "(none)",
		       ReqDocName,
		       Options->TraceDesignState.HighLevelRequirementSection.Number,
		       DesDocName,
		       &Text[pmatch[0].rm_so],
		       DesDocName,
		       Options->DocSections[DES_DOC].Number);
#endif		       
	    }
	  Text[pmatch[0].rm_eo] = c;
	}
      for (; !i && !regexec (&Options->LowSectionPatternMachine, Text, 1, pmatch, 0); i++)
        {
	  printf (" %s", Options->DocSections[DES_DOC].Number);
	  for (k = 0; k < Options->TraceDesignState.NumHighLevelRequirements; k++)
	    {
#ifndef TRACE_DESIGN_FULL
	      fprintf (Options->FileTraceDesign, "\"%s %s\",\"%s %s\"\n",
		       ReqDocName,
		       Options->TraceDesignState.HighLevelRequirements[k],
		       DesDocName,
		       Options->DocSections[DES_DOC].Number);
#else		       
	      fprintf (Options->FileTraceDesign, "\"%s %s\",\"%s %s\",\"%s %s\",\"%s %s\"\n",
		       ReqDocName,
		       Options->TraceDesignState.HighLevelRequirements[k],
		       ReqDocName,
		       Options->TraceDesignState.HighLevelRequirementSection.Number,
		       DesDocName,
		       Options->DocSections[DES_DOC].Number,
		       DesDocName,
		       Options->DocSections[DES_DOC].Number);
#endif		       
	    }
	  if (k == 0)
	    {
#ifndef TRACE_DESIGN_FULL
	      fprintf (Options->FileTraceDesign, "\"%s\",\"%s %s\"\n",
		       "(none)",
		       DesDocName,
		       Options->DocSections[DES_DOC].Number);
#else		      
	      fprintf (Options->FileTraceDesign, "\"%s\",\"%s %s\",\"%s %s\",\"%s %s\"\n",
		       "(none)",
		       ReqDocName,
		       Options->TraceDesignState.HighLevelRequirementSection.Number,
		       DesDocName,
		       Options->DocSections[DES_DOC].Number,
		       DesDocName,
		       Options->DocSections[DES_DOC].Number);
#endif		       
	    }
	}
      if (i == 0)
        {
          printf (" (none)");
#ifdef TRACE_DESIGN_FULL
          for (k = 0; k < Options->TraceDesignState.NumHighLevelRequirements; k++)
	    fprintf (Options->FileTraceDesign, "\"%s %s\",\"%s %s\",\"%s\",\"%s %s\"\n",
		     ReqDocName,
		     Options->TraceDesignState.HighLevelRequirements[k],
		     ReqDocName,
		     Options->TraceDesignState.HighLevelRequirementSection.Number,
		     "(none)",
		     DesDocName,
		     Options->DocSections[DES_DOC].Number);
	  if (k == 0)
	    fprintf (Options->FileTraceDesign, "\"%s\",\"%s %s\",\"%s\",\"%s %s\"\n",
		     "(none)",
		     ReqDocName,
		     Options->TraceDesignState.HighLevelRequirementSection.Number,
		     "(none)",
		     DesDocName,
		     Options->DocSections[DES_DOC].Number);
#endif		       
	}
      printf ("\n");
      
      Options->TraceDesignState.LowLevelCount += i;
      
      xmlFree(Text);      
    }

  return (0);
}


